//
//			Yasuharu Okamoto, 2012.05, 2012.06
//
/************************************************

      R. L. Eubank and A. Kupresanin
	  "Statistical Computing in C++ and R", 2012, P. 129.

	  The period is approximately 2**191 (about 3 * 10**57)
	  if
	    x_1_0, x_1_1, x_1_2 are not all 0,
	    and all are less than 2**32 - 209,
      and 
	    x_2_0, x_2_1, x_2_1 are not all 0,
	    and all are less than 2**32 - 22853.

************************************************/

#include <cmath>

namespace rnp2to191 {

	class RNP2to191 {
		public:
				int ic, ib;

				long int x1_0, x1_1, x1_2, x1_3;
				long int x2_0, x2_1, x2_2, x2_3;
				long int c2p32minus209, c2p32minus22853;
				double inv2p32minus208, cforzeroZn;

				inline long int DivModP( long int a, long int m ){
						long int v = a % m;
						if (v < 0) v += m;
						return v;
				}

				RNP2to191(){
					ic = ib = 0;
					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;
				}

				RNP2to191( int icycle, int ibias ){
					ic = icycle;
					ib = ibias;

					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;

					if (ib > 0)
						for (int i = 0; i < ib; i++) uni0();
				}

				RNP2to191( long int a1, long int a2, long int a3, long int a4, long int a5, long int a6 ){
					ic = ib = 0;

					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = a1;
					x1_1 = a2;
					x1_0 = a3;
					x2_2 = a4;
					x2_1 = a5;
					x2_0 = a6;
				}

				inline void next(){
					x1_3 = x1_2; 
					x1_2 = x1_1;
					x1_1 = x1_0;
					x2_3 = x2_2;
					x2_2 = x2_1;
					x2_1 = x2_0;
					x1_0 = DivModP( 1403580 * x1_2 - 810728 * x1_3, c2p32minus209 );
					x2_0 = DivModP( 527612 * x2_1 - 1370589 * x2_3, c2p32minus22853 );
		
					return;
				}

				double uni0(){
					next();
					long int zn = DivModP( x1_0 - x2_0, c2p32minus209 );
					if (zn > 0){
							return zn * inv2p32minus208;
					}else{
							return cforzeroZn;
					}
				}

				void Init(){
					ic = ib = 0;

					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;
				}

				void Init( long int a1, long int a2, long int a3, long int a4, long int a5, long int a6 ){
					ic = ib = 0;

					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = a1;
					x1_1 = a2;
					x1_0 = a3;
					x2_2 = a4;
					x2_1 = a5;
					x2_0 = a6;
				}

				void Init( int icycle, int ibias ){
					ic = icycle;
					ib = ibias;

					c2p32minus209 = 0x0000000100000000L - 209;
					c2p32minus22853 = 0x0000000100000000L - 22853;
					inv2p32minus208 = 1.0 / (pow(2.0, 32.0) - 208.0);
					cforzeroZn = (pow(2.0, 32.0) - 209.0) / (pow(2.0, 32.0) - 208.0);
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;

					if (ib > 0)
						for (int i = 0; i < ib; i++) next();
				}

				double uni(){
					for (int i = 0; i < ic - 1; i++) next();

					next();
					long int zn = DivModP( x1_0 - x2_0, c2p32minus209 );
					if (zn > 0){
							return zn * inv2p32minus208;
					}else{
							return cforzeroZn;
					}
				}

				
    //        Standard Normal Distribution
	//        Rejection polar method for normal variate
	double normal();


    //    Normal Distribution of mean m  and standard deviation s
	double normalMS(double m, double s);


    //    A Pair of Independent Random Numbers of 
	//      the Standard Normal Distribution 
	void  normalPair( double & n1, double & n2 );


	};

}
